/*
 * This file is autogenerated by autogen.pl. Do not edit this file by hand.
 */
#ifndef ORTE_FRAMEWORKS_H
#define ORTE_FRAMEWORKS_H

#include <opal/mca/base/mca_base_framework.h>

extern mca_base_framework_t orte_errmgr_base_framework;
extern mca_base_framework_t orte_ess_base_framework;
extern mca_base_framework_t orte_filem_base_framework;
extern mca_base_framework_t orte_grpcomm_base_framework;
extern mca_base_framework_t orte_iof_base_framework;
extern mca_base_framework_t orte_odls_base_framework;
extern mca_base_framework_t orte_oob_base_framework;
extern mca_base_framework_t orte_plm_base_framework;
extern mca_base_framework_t orte_ras_base_framework;
extern mca_base_framework_t orte_regx_base_framework;
extern mca_base_framework_t orte_rmaps_base_framework;
extern mca_base_framework_t orte_rml_base_framework;
extern mca_base_framework_t orte_routed_base_framework;
extern mca_base_framework_t orte_rtc_base_framework;
extern mca_base_framework_t orte_schizo_base_framework;
extern mca_base_framework_t orte_snapc_base_framework;
extern mca_base_framework_t orte_sstore_base_framework;
extern mca_base_framework_t orte_state_base_framework;

static mca_base_framework_t *orte_frameworks[] = {
    &orte_errmgr_base_framework,
    &orte_ess_base_framework,
    &orte_filem_base_framework,
    &orte_grpcomm_base_framework,
    &orte_iof_base_framework,
    &orte_odls_base_framework,
    &orte_oob_base_framework,
    &orte_plm_base_framework,
    &orte_ras_base_framework,
    &orte_regx_base_framework,
    &orte_rmaps_base_framework,
    &orte_rml_base_framework,
    &orte_routed_base_framework,
    &orte_rtc_base_framework,
    &orte_schizo_base_framework,
    &orte_snapc_base_framework,
    &orte_sstore_base_framework,
    &orte_state_base_framework,
    NULL
};

#endif /* ORTE_FRAMEWORKS_H */

