/*
 * Copyright (c) 2013      Mellanox Technologies, Inc.
 *                         All rights reserved.
 * Copyright (c) 2016      Research Organization for Information Science
 *                         and Technology (RIST). All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
#ifndef SHMEM_TYPES_H
#define SHMEM_TYPES_H


/*
 * Predefine some internal types so we dont need all the include
 * dependencies.
 */

   struct ompi_proc_t;
   struct oshmem_group_t;
   struct oshmem_op_t;

#endif
