/*
 * Copyright (c) 2010      Cisco Systems, Inc.  All rights reserved.
 * Copyright (c) 2016      Intel, Inc.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 *
 */

#ifndef PMIX_PIF_BASE_H
#define PMIX_PIF_BASE_H

#include "pmix_config.h"

#include "src/mca/base/pmix_mca_base_framework.h"

#include "src/mca/pif/pif.h"

BEGIN_C_DECLS

/*
 * Framework declaration
 */
PMIX_EXPORT extern pmix_mca_base_framework_t pmix_pif_base_framework;

END_C_DECLS

#endif /* PMIX_BASE_PIF_H */
