! -*- fortran -*-
! $HEADER$
!
! *** THIS FILE IS AUTOMATICALLY GENERATED!
! *** Any manual edits will be lost!
!
module mpi_ext
!     Even though this is not a useful parameter (cannot be used as a
!     preprocessor catch) define it to keep the linker from complaining
!     during the build.
      integer OMPI_HAVE_MPI_EXT
      parameter (OMPI_HAVE_MPI_EXT=1)
!
!
!     Enabled Extension: affinity
!     No "use mpi" bindings available
!

!
!     Enabled Extension: cuda
!     No "use mpi" bindings available
!

!
end module mpi_ext
