/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
/*
 * Copyright (c) 2007-2015 Los Alamos National Security, LLC. All rights
 *                         reserved.
 * Copyright (c) 2004-2008 The Trustees of Indiana University.
 *                         All rights reserved.
 * Copyright (c) 2016      Intel, Inc.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#include "orte_config.h"
#include "orte/constants.h"

#include "opal/mca/base/base.h"

#include "orte/mca/routed/base/base.h"
#include "routed_binomial.h"

static int orte_routed_binomial_component_query(mca_base_module_t **module, int *priority);

/**
 * component definition
 */
orte_routed_component_t mca_routed_binomial_component = {
    /* First, the mca_base_component_t struct containing meta
       information about the component itself */

    .base_version = {
        ORTE_ROUTED_BASE_VERSION_3_0_0,

        .mca_component_name = "binomial",
        MCA_BASE_MAKE_VERSION(component, ORTE_MAJOR_VERSION, ORTE_MINOR_VERSION,
                              ORTE_RELEASE_VERSION),
        .mca_query_component = orte_routed_binomial_component_query
    },
    .base_data = {
        /* This component can be checkpointed */
        MCA_BASE_METADATA_PARAM_CHECKPOINT
    },
};

static int orte_routed_binomial_component_query(mca_base_module_t **module, int *priority)
{
    /* make this selected ONLY if the user directs as this module scales
     * poorly compared to our other options
     */
    *priority = 30;
    *module = (mca_base_module_t *) &orte_routed_binomial_module;
    return ORTE_SUCCESS;
}
