/*
 * Copyright (c) 2013      Mellanox Technologies, Inc.
 *                         All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 */

#ifndef MCA_MEMHEAP_BUDDY_COMPONENT_H
#define MCA_MEMHEAP_BUDDY_COMPONENT_H

BEGIN_C_DECLS

/*
 * MEMHEAP module functions.
 */
OSHMEM_MODULE_DECLSPEC extern mca_memheap_base_component_2_0_0_t mca_memheap_buddy_component;

END_C_DECLS

#endif
