/*
 * Copyright (c) 2018      Mellanox Technologies, Inc.
 *                         All rights reserved.
 * Copyright (c) 2020      Intel, Inc.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef GDS_DS21_FILE_H
#define GDS_DS21_FILE_H

#include "src/include/pmix_config.h"
#include "include/pmix_common.h"

extern pmix_common_dstore_file_cbs_t pmix_ds21_file_module;

#endif // GDS_DS21_FILE_H
