/*
 * Copyright (c) 2004-2007 The Trustees of the University of Tennessee.
 *                         All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef PML_V_H_HAS_BEEN_INCLUDED
#define PML_V_H_HAS_BEEN_INCLUDED

#include "ompi_config.h"
#include "ompi/mca/pml/pml.h"
#include "ompi/request/request.h"

#include "ompi/mca/vprotocol/base/base.h"

BEGIN_C_DECLS

OMPI_MODULE_DECLSPEC extern mca_pml_base_component_2_0_0_t mca_pml_v_component;

END_C_DECLS

#endif  /* PML_V_H_HAS_BEEN_INCLUDED */
