/*
 * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2013 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2012-2013 Los Alamos National Security, Inc.  All rights reserved.
 * Copyright (c) 2012-2013 Inria.  All rights reserved.
 * Copyright (c) 2014      Research Organization for Information Science
 *                         and Technology (RIST). All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#include "ompi_config.h"

#include <stdio.h>

#include "ompi/constants.h"
#include "opal/class/opal_list.h"
#include "opal/util/output.h"
#include "ompi/mca/mca.h"
#include "opal/mca/base/base.h"


#include "ompi/mca/topo/base/base.h"

/*
 * The static-component.h is generated by the configure script. It contains
 * statements and the definition of an array of pointers to each component's
 * public mca_base_component_t struct.
 */
#include "ompi/mca/topo/base/static-components.h"

static void mca_topo_base_module_construct(mca_topo_base_module_t * topo) {
    memset(&(topo->mtc), 0, sizeof(topo->mtc));
}

static void mca_topo_base_module_destruct(mca_topo_base_module_t * topo) {
    /* topo->mtc is an union of pointers to opal_object_t.
       In order to release it, we just have to call OBJ_RELEASE on any of the member,
       (cart in this case) and the appropriate object destructor will be called */
    if (NULL != topo->mtc.cart) {
        OBJ_RELEASE(topo->mtc.cart);
    }
}

OBJ_CLASS_INSTANCE(mca_topo_base_module_t, opal_object_t,
                   mca_topo_base_module_construct,
                   mca_topo_base_module_destruct);

static int mca_topo_base_close(void)
{
    return mca_base_framework_components_close(&ompi_topo_base_framework, NULL);
}

/**
 * Function for finding and opening either all the MCA topo components, or
 * the one that specifically requested via a MCA parameter.
 */
static int mca_topo_base_open(mca_base_open_flag_t flags)
{
    return mca_base_framework_components_open(&ompi_topo_base_framework, flags);
}

int mca_topo_base_neighbor_count (ompi_communicator_t *comm, int *indegree, int *outdegree) {
  if (!OMPI_COMM_IS_TOPO(comm)) {
    return OMPI_ERR_BAD_PARAM;
  }

  if (OMPI_COMM_IS_CART(comm)) {
    /* cartesian */
    /* outdegree is always 2*ndims because we need to iterate over
       empty buffers for MPI_PROC_NULL */
    *outdegree = *indegree = 2 * comm->c_topo->mtc.cart->ndims;
  } else if (OMPI_COMM_IS_GRAPH(comm)) {
    /* graph */
    int rank, nneighbors;

    rank = ompi_comm_rank (comm);
    mca_topo_base_graph_neighbors_count (comm, rank, &nneighbors);

    *outdegree = *indegree = nneighbors;
  } else if (OMPI_COMM_IS_DIST_GRAPH(comm)) {
    /* graph */
    *indegree = comm->c_topo->mtc.dist_graph->indegree;
    *outdegree = comm->c_topo->mtc.dist_graph->outdegree;
  }

  return OMPI_SUCCESS;
}

MCA_BASE_FRAMEWORK_DECLARE(ompi, topo, "OMPI Topo", NULL,
                           mca_topo_base_open, mca_topo_base_close,
                           mca_topo_base_static_components, 0);

