/*
 * Copyright (c) 2004-2008 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2009 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2011      Los Alamos National Security, LLC.  All rights
 *                         reserved.
 * Copyright (c) 2008      Sun Microsystems, Inc.  All rights reserved.
 * Copyright (c) 2011      IBM Corporation.  All rights reserved.
 * Copyright (c) 2014      Intel, Inc.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef ORTE_PLM_ISOLATED_EXPORT_H
#define ORTE_PLM_ISOLATED_EXPORT_H

#include "orte_config.h"

#include "orte/mca/mca.h"

#include "orte/mca/plm/plm.h"

BEGIN_C_DECLS


ORTE_MODULE_DECLSPEC extern orte_plm_base_component_t mca_plm_isolated_component;
extern orte_plm_base_module_t orte_plm_isolated_module;


END_C_DECLS

#endif /* ORTE_PLS_ISOLATED_EXPORT_H */
