/*
 * Copyright (c) 2010      Cisco Systems, Inc.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 *
 */

#ifndef OPAL_IF_BASE_H
#define OPAL_IF_BASE_H

#include "opal_config.h"

#include "opal/mca/base/mca_base_framework.h"

#include "opal/mca/if/if.h"

BEGIN_C_DECLS

/*
 * Framework declaration
 */
OPAL_DECLSPEC extern mca_base_framework_t opal_if_base_framework;

END_C_DECLS

#endif /* OPAL_BASE_IF_H */
