/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 * Copyright (c) 2011      Los Alamos National Security, LLC.  All
 *                         right reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef BTL_OPENIB_CONNECT_UD_H
#define BTL_OPENIB_CONNECT_UD_H

#include "opal_config.h"

#include "connect/connect.h"

extern opal_btl_openib_connect_base_component_t opal_btl_openib_connect_udcm;

#endif
