/*
 * Copyright (c) 2009-2012 Mellanox Technologies.  All rights reserved.
 * Copyright (c) 2009-2012 Oak Ridge National Laboratory.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef COMM_NETPATTERNS_H
#define COMM_NETPATTERNS_H

#include "ompi_config.h"

BEGIN_C_DECLS

#define MAX_TMP_BUFFER            8192

END_C_DECLS

#endif /* COMM_NETPATTERNS_H */
