/*
 * Copyright (c) 2011-2017 Cisco Systems, Inc.  All rights reserved
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef ORTE_RMAPS_PPR_H
#define ORTE_RMAPS_PPR_H

#include "orte_config.h"

#include "opal/mca/hwloc/hwloc-internal.h"

#include "orte/mca/rmaps/rmaps.h"

BEGIN_C_DECLS

ORTE_MODULE_DECLSPEC extern orte_rmaps_base_component_t mca_rmaps_ppr_component;
extern orte_rmaps_base_module_t orte_rmaps_ppr_module;


END_C_DECLS

#endif
