/*
 * Copyright (c) 2007-2008 Cisco Systems, Inc.  All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef BTL_OPENIB_CONNECT_RDMACM_H
#define BTL_OPENIB_CONNECT_RDMACM_H

#include "opal_config.h"

#include "connect/connect.h"

extern opal_btl_openib_connect_base_component_t opal_btl_openib_connect_rdmacm;

#endif
